package http

import (
	"bytes"
	"encoding/json"
	"fmt"
	"io"
	"net/http/httptest"
	"testing"
	"time"

	"github.com/gogo/protobuf/jsonpb"
	"github.com/gogo/protobuf/proto"
	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
	xbus "xbus.io/go-xbus/v3"
	"xbus.io/go-xbus/v3/api"

	fullenv "xbus.io/xbus/fullenv_tests"
	"xbus.io/xbus/testutils"
)

type LoggingReader struct {
	reader io.Reader
	tb     testing.TB
}

func (r LoggingReader) Read(p []byte) (n int, err error) {
	n, err = r.reader.Read(p)
	r.tb.Logf("Just read %s", string(p))
	return
}

func setupEnv(t *testing.T, filename string, startup bool) (env *fullenv.FullEnv, gw *Gateway) {
	t.Helper()

	defer func() {
		if gw == nil {
			env.Close()
		}
	}()
	logger := testutils.GetTestLogger(t)
	options, err := fullenv.OptionsFromFile(filename)
	options.Logger = logger
	require.NoError(t, err)
	env = fullenv.NewFullEnv("xbus-http-emitter", options)
	require.NoError(t, env.Init())

	if startup {
		require.NoError(t, env.Startup())
	}

	coptions := xbus.COptions{
		Name: "gw",
		Type: api.Account_GATEWAY,
	}
	client := xbus.NewClient(coptions, logger)
	require.NoError(t, client.CreateKey(false))
	require.NoError(t, client.CreateCSR("localhost", "xbus", "xbus", false))
	require.NoError(t, client.Connect())
	s, err := client.Register()
	require.NoError(t, err)
	if s != api.Registration_PENDING {
		t.Fatal(fmt.Sprintf("Invalid state: %s", s))
	}
	require.NoError(t, env.XbusCtl.Run("account", "accept", "gw"))
	s, err = client.Register()
	require.NoError(t, err)
	if s != api.Registration_ACCEPTED {
		t.Fatal(fmt.Sprintf("Invalid state: %s", s))
	}
	client.Close()
	coptions = client.GetOptions()
	gw = NewGateway(coptions, logger)
	return
}

func TestEmitter(t *testing.T) {
	for _, encoding := range []string{"json", "proto"} {
		t.Run(encoding, func(t *testing.T) {
			env, gateway := setupEnv(t, "emitter-test.yaml", true)
			defer env.Close()

			apiKey, err := env.XbusCtl.RunReadLastLine("account", "renew-apikey", "outerclient")
			require.NoError(t, err)

			var buf io.Reader
			if encoding == "json" {
				buf = bytes.NewBufferString(`{
		"output": "default",
		"envelope": {
			"event_ids": ["53131466-4e2e-11e7-916e-af2b1e6d242d"],
			"events": [{
				"id": "53131466-4e2e-11e7-916e-af2b1e6d242d",
				"type": "test",
				"ItemCount": 1,
				"Items": ["QSBtZXNzYWdl"]
			}]
		}
	}`)
			} else {
				b, err := proto.Marshal(&api.OutputRequest{
					Output: "default",
					Envelope: &api.Envelope{
						EventIDs: []api.UUID{
							api.MustUUIDFromString("53131466-4e2e-11e7-916e-af2b1e6d242d"),
						},
						Events: []api.Event{
							{ID: api.MustUUIDFromString("53131466-4e2e-11e7-916e-af2b1e6d242d"),
								Type:      "test",
								ItemCount: 1,
								Items:     []api.Item{api.Item("A message")},
							},
						},
					},
				})
				if err != nil {
					t.Fatal(err)
				}
				buf = bytes.NewBuffer(b)
			}

			req := httptest.NewRequest("POST", "/emitter/output", buf)
			if encoding == "json" {
				req.Header.Set("Content-Type", "application/json")
			} else {
				req.Header.Set("Content-Type", "application/x-protobuf")
			}
			req.Header.Set("Xbus-API-Key", apiKey)

			w := httptest.NewRecorder()

			gateway.Router.ServeHTTP(w, req)

			if !assert.Equal(t, 200, w.Code) {
				t.Fatalf("Body is: %s", w.Body.String())
			}

			if w.Body.String() != "OK" {
				t.Errorf("Unexpected response: %s", w.Body.Bytes())
				return
			}

			time.Sleep(time.Second)

			assert.Equal(t, 1, len(env.Consumer.Envelopes))

			e := env.Consumer.Envelopes[0]
			assert.Equal(t, "A message", string(e.Events[0].Items[0]))
			assert.Equal(t, "53131466-4e2e-11e7-916e-af2b1e6d242d", e.Events[0].ID.String())
		})

		t.Run(encoding+"/ID-less event", func(t *testing.T) {
			env, gateway := setupEnv(t, "emitter-test.yaml", true)
			defer env.Close()

			apiKey, err := env.XbusCtl.RunReadLastLine("account", "renew-apikey", "outerclient")
			require.NoError(t, err)

			var buf io.Reader
			if encoding == "json" {
				buf = bytes.NewBufferString(`{
		"output": "default",
		"envelope": {
			"events": [{
				"type": "test",
				"ItemCount": 1,
				"Items": ["QSBtZXNzYWdl"]
			}]
		}
	}`)
			} else {
				b, err := proto.Marshal(&api.OutputRequest{
					Output: "default",
					Envelope: &api.Envelope{
						Events: []api.Event{
							{
								Type:      "test",
								ItemCount: 1,
								Items:     []api.Item{api.Item("A message")},
							},
						},
					},
				})
				if err != nil {
					t.Fatal(err)
				}
				buf = bytes.NewBuffer(b)
			}

			req := httptest.NewRequest("POST", "/emitter/output", buf)
			if encoding == "json" {
				req.Header.Set("Content-Type", "application/json")
			} else {
				req.Header.Set("Content-Type", "application/x-protobuf")
			}
			req.Header.Set("Xbus-API-Key", apiKey)

			w := httptest.NewRecorder()

			gateway.Router.ServeHTTP(w, req)

			if !assert.Equal(t, 200, w.Code) {
				t.Fatalf("Body is: %s", w.Body.String())
			}

			if w.Body.String() != "OK" {
				t.Errorf("Unexpected response: %s", w.Body.Bytes())
				return
			}

			time.Sleep(time.Second)

			assert.Equal(t, 1, len(env.Consumer.Envelopes))

			e := env.Consumer.Envelopes[0]
			assert.Equal(t, "A message", string(e.Events[0].Items[0]))
		})
	}
}

func TestConsumer(t *testing.T) {
	env, gateway := setupEnv(t, "consumer-test.yaml", true)
	defer env.Close()

	apiKey, err := env.XbusCtl.RunReadLastLine("account", "renew-apikey", "outerclient")
	require.NoError(t, err)

	req := httptest.NewRequest("GET", "/test-consumer/processrequest", nil)
	req.Header.Set("Xbus-API-Key", apiKey)
	req.Header.Set("Accept", "application/json")
	req.Header.Set("Accept-Item-Encoding", "json,text,binary")

	w := httptest.NewRecorder()

	sender := env.Emitter.Actor.OpenOutput(nil, "default", "test")
	require.NoError(t, sender.AddItems("test", api.Item("a item")))
	require.NoError(t, sender.Close())

	envStateChan, sub, err := env.Emitter.Actor.GetClient().
		API.ProcessState.EnvelopeStatesSubscribeChan(
		env.Emitter.Actor.ID.String(),
		sender.GetID().String())
	if err != nil {
		t.Fatal(err)
	}
	defer testutils.DeferedAssertNoError(t, sub.Unsubscribe)

	done := make(chan bool)

	go func() {
		gateway.Router.ServeHTTP(w, req)
		done <- true
	}()

	select {
	case <-time.After(10 * time.Second):
		t.Fatal("Timeout...")
	case <-done:
		t.Log("Request served, checking result")
	}

	if !assert.Equal(t, 200, w.Code) {
		t.Fatalf("Body is: %s", w.Body.String())
	}

	dec := json.NewDecoder(LoggingReader{w.Body, t})

	var reply api.ActorProcessRequest

	if err := jsonpb.UnmarshalNext(dec, &reply); err != nil {
		t.Fatalf("Error decoding response: %s", err)
	}
	envelope := reply.Inputs[0].Envelope
	if len(envelope.Events) != 1 {
		t.Fatalf("Expected 1 event, got %d", len(envelope.Events))
	}
	event := envelope.Events[0]
	if len(event.Items) != 1 {
		t.Fatalf("Expected 1 item, got %d", len(event.Items))
	}
	item := event.Items[0]
	assert.Equal(t, "a item", string(item))

	// TODO decode the envelope and check its items

	{
		buf := bytes.NewBuffer(nil)
		enc := json.NewEncoder(buf)
		if err := enc.Encode(map[string]interface{}{
			"context": reply.Context,
			"status":  "SUCCESS",
		}); err != nil {
			t.Fatal(err)
		}

		req := httptest.NewRequest("POST", "/test-consumer/processingend", buf)
		req.Header.Set("Xbus-API-Key", apiKey)

		w := httptest.NewRecorder()

		done := make(chan bool)

		go func() {
			gateway.Router.ServeHTTP(w, req)
			done <- true
		}()

		select {
		case <-time.After(10 * time.Second):
			t.Fatal("Timeout...")
		case <-done:
			t.Log("Request served, checking result")
		}

		if !assert.Equal(t, 200, w.Code) {
			t.Fatalf("Body is: %s", w.Body.String())
		}
	}
	timeout := time.After(5 * time.Second)
	for {
		select {
		case <-timeout:
			t.Errorf("Process state not updated to 'done'")
			return
		case envState := <-envStateChan:
			if envState.Status == api.Process_DONE {
				return
			}
		}
	}
}

type testReplier struct {
}

func (r testReplier) Process(apc *xbus.ActorProcessingContext) error {
	env, err := apc.CompleteEnvelope("request", time.Second)
	if err != nil {
		return err
	}
	out := apc.OpenOutput("response", "test.response")
	if err := out.AddItems(
		"test.response",
		api.Item(string(env.Events[0].Items[0])+" reply"),
	); err != nil {
		return err
	}
	if err := out.Close(); err != nil {
		return err
	}
	return nil
}

func TestReply(t *testing.T) {
	for _, encoding := range []string{"json", "proto"} {
		t.Run(encoding, func(t *testing.T) {
			env, gateway := setupEnv(t, "reply-test.yaml", false)
			defer env.Close()

			env.RegisterActorService("replier", func(*xbus.Actor) xbus.ActorService {
				return testReplier{}
			})

			if err := env.Startup(); err != nil {
				t.Fatal(err)
			}

			apiKey, err := env.XbusCtl.RunReadLastLine("account", "renew-apikey", "outerclient")
			require.NoError(t, err)

			var buf io.Reader
			if encoding == "json" {
				buf = bytes.NewBufferString(`{
		"envelope": {
			"id": "53131466-4e2e-11e7-916e-af2b1e6d242e",
			"eventIDs": ["53131466-4e2e-11e7-916e-af2b1e6d242d"],
			"events": [{
				"id": "53131466-4e2e-11e7-916e-af2b1e6d242d",
				"encoding": "text",
				"type": "test-request",
				"ItemCount": 1,
				"Items": ["A message"]
			}]
		}
	}`)
			} else {
				b, err := proto.Marshal(&api.OutputRequest{
					Envelope: &api.Envelope{
						EventIDs: []api.UUID{
							api.MustUUIDFromString("53131466-4e2e-11e7-916e-af2b1e6d242d"),
						},
						Events: []api.Event{
							{ID: api.MustUUIDFromString("53131466-4e2e-11e7-916e-af2b1e6d242d"),
								Type:      "test-request",
								ItemCount: 1,
								Items:     []api.Item{api.Item("A message")},
							},
						},
					},
				})
				if err != nil {
					t.Fatal(err)
				}
				buf = bytes.NewBuffer(b)
			}

			req := httptest.NewRequest("POST", "/requestor/request", buf)
			if encoding == "json" {
				req.Header.Set("Content-Type", "application/json")
			} else {
				req.Header.Set("Content-Type", "application/x-protobuf")
			}
			req.Header.Set("Xbus-API-Key", apiKey)

			w := httptest.NewRecorder()

			gateway.Router.ServeHTTP(w, req)

			if !assert.Equal(t, 200, w.Code) {
				t.Fatalf("Body is: %s", w.Body.String())
			}

			var response api.Envelope

			if encoding == "json" {
				err = jsonpb.Unmarshal(w.Body, &response)
			} else {
				err = proto.Unmarshal(w.Body.Bytes(), &response)
			}

			if err != nil {
				t.Fatal(err)
			}

			assert.Equal(t, "A message reply", string(response.Events[0].Items[0]))
		})
	}
}
