package director

import (
	"context"

	"github.com/nats-io/nats.go"
	"github.com/rs/zerolog"
	"xbus.io/go-xbus/v3/api"

	"xbus.io/xbus/process"
	svc_tools "xbus.io/xbus/services/tools"
	"xbus.io/xbus/storage"
)

// ActorSetConfig holds the settings for instanciating an ActorSet
type ActorSetConfig struct {
	Log             zerolog.Logger
	Conn            *nats.Conn
	ActorStorage    storage.ActorStorage
	JobStorage      storage.JobStorage
	ProcessManager  *process.Manager
	EnvelopeStorage storage.EnvelopeStorage
	ActorMap        *svc_tools.ActorMap
	Director        directorServer
	EnvelopeWatcher *EnvelopeWatcher
}

// ActorSet
// This API is NOT thread-safe
type ActorSet struct {
	ctx context.Context
	cfg ActorSetConfig

	log zerolog.Logger

	handlers map[api.UUID]*ActorHandler
}

// newActorSet instanciates a ActorSet
func newActorSet(ctx context.Context, config ActorSetConfig) *ActorSet {
	return &ActorSet{
		ctx:      ctx,
		cfg:      config,
		handlers: make(map[api.UUID]*ActorHandler),
		log:      config.Log,
	}
}

// Submit dispatches a job to the correct actor handler if it exists
func (s *ActorSet) Submit(job *Job) {
	handler, ok := s.handlers[job.Node.ActorID]
	// If a handler exists, add the job to it
	if ok {
		handler.Submit(job)
	}
}

// CancelProcessJobs cancel all running jobs that belongs to the given process
func (s *ActorSet) CancelProcessJobs(processID api.UUID) {
	for _, handler := range s.handlers {
		handler.CancelProcessJobs(processID)
	}
}

// Get returns the handler of a given actor, or nil if it does not exist
func (s *ActorSet) Get(id api.UUID) *ActorHandler {
	return s.handlers[id]
}

// GetOrCreate returns the handler of a given actor, creating it if needed
func (s *ActorSet) GetOrCreate(id api.UUID) *ActorHandler {
	// init the queue if needed
	handler, ok := s.handlers[id]
	if !ok {
		actor, ok := s.cfg.ActorStorage.Get(id)
		if !ok {
			s.log.Error().Str("actor-id", id.String()).Msg("unknown actor")
			return nil
		}
		handler = newActorHandler(
			s.ctx,
			actor,
			ActorHandlerConfig{
				Log:             s.log.With().Str("actor-id", id.String()).Logger(),
				Conn:            s.cfg.Conn,
				ActorStorage:    s.cfg.ActorStorage,
				JobStorage:      s.cfg.JobStorage,
				ProcessManager:  s.cfg.ProcessManager,
				EnvelopeStorage: s.cfg.EnvelopeStorage,
				ActorMap:        s.cfg.ActorMap,
				Director:        s.cfg.Director,
				EnvelopeWatcher: s.cfg.EnvelopeWatcher,
			},
		)
		s.handlers[id] = handler
	}
	return handler
}

// SetOnline starts the handler of the given actor
func (s *ActorSet) SetOnline(id api.UUID) {
	if handler := s.GetOrCreate(id); handler != nil {
		handler.Start()
	}
}

// SetOnline stops the handler of the given actor
func (s *ActorSet) SetOffline(id api.UUID) {
	if handler := s.Get(id); handler != nil {
		handler.Stop()
	}
}

// Close closes all the underlying handlers
func (s *ActorSet) Close() {
	for _, handler := range s.handlers {
		handler.Close()
	}
}
