// Copyright 2017 Orus.io. All rights reserved.

package sql

import (
	"time"

	"github.com/m4rw3r/uuid"
	"github.com/orus-io/yago"
	"github.com/rs/zerolog"
	"github.com/slicebit/qb"
	"xbus.io/go-xbus/v3/api"
	"xbus.io/go-xbus/v3/api/control"

	"xbus.io/xbus/storage/sql/model"
	xuuid "xbus.io/xbus/uuid"
)

// NewLogStorage initialize a LogStorage
func NewLogStorage(model *model.Model, db *yago.DB, logger zerolog.Logger) LogStorage {
	return LogStorage{Base{
		model: model,
		db:    db,
		logger: logger.With().
			Str("storage", "log").
			Str("implementation", "sql").
			Logger(),
	}}
}

// LogStorage implements storage.LogStorage
type LogStorage struct {
	Base
}

// Log adds a log entry to the storage
func (s LogStorage) Log(entry api.LogEntry) {
	tx := s.MustBegin()
	defer tx.RollbackIfOpened()

	var dbLogEntry = model.LogEntry{
		EnvelopeID: xuuid.NullUUID(entry.EnvelopeID),
		ActorID:    xuuid.NullUUID(entry.ActorID),
		ProcessID:  xuuid.NullUUID(entry.ProcessID),
		NodeID:     entry.NodeID,
		Time:       entry.Message.Time.UTC(),
		Level:      entry.Message.Level,
		Text:       entry.Message.Text,
	}
	if err := tx.Insert(&dbLogEntry); err != nil {
		panic(err)
	}
	tx.MustCommit()
}

func (s LogStorage) list(crits ...qb.Clause) []api.LogEntry {
	tx := s.MustBegin()
	defer tx.BlindRollback()

	var dbEntries []model.LogEntry
	if err := tx.Query(s.model.LogEntry).Where(crits...).All(&dbEntries); err != nil {
		panic(err)
	}

	var entries = make([]api.LogEntry, len(dbEntries))
	for i, dbEntry := range dbEntries {
		entries[i] = api.LogEntry{
			EnvelopeID: api.UUID(dbEntry.EnvelopeID),
			ActorID:    api.UUID(dbEntry.ActorID),
			ProcessID:  api.UUID(dbEntry.ProcessID),
			NodeID:     dbEntry.NodeID,
			Message: api.LogMessage{
				Time:  dbEntry.Time,
				Level: dbEntry.Level,
				Text:  dbEntry.Text,
			},
		}
	}
	return entries
}

// GetProcessMessages returns all the messages related to the given envelope
func (s LogStorage) GetProcessMessages(processID api.UUID, level api.LogLevel) []api.LogEntry {
	var crits = []qb.Clause{
		s.model.LogEntry.ProcessID.Eq(uuid.UUID(processID)),
	}

	switch level {
	case api.LogLevel_ERROR:
		crits = append(crits, s.model.LogEntry.Level.Eq(level))
	case api.LogLevel_WARNING:
		crits = append(crits, s.model.LogEntry.Level.NotEq(api.LogLevel_NOTICE))
	}

	return s.list(crits...)
}

// PurgeBefore removes all logs older than the given date
func (s LogStorage) PurgeBefore(t time.Time) int64 {
	tx := s.MustBegin()
	defer tx.MustCommit()

	t = t.UTC()

	res, err := tx.GetEngine().Exec(
		s.model.LogEntry.GetMapper().Table().Delete().Where(
			s.model.LogEntry.Time.Lt(t)))
	must(err)
	ra, err := res.RowsAffected()
	must(err)
	return ra
}

// Stat returns a summary of the storage content
func (s LogStorage) Stat() (*control.StorageStat_Entry, error) {
	tx := s.MustBegin()
	defer tx.BlindRollback()

	var stat = control.StorageStat_Entry{
		Name:           "default",
		Type:           control.StorageType_LOG,
		Implementation: "pg",
	}

	if err := tx.GetEngine().
		QueryRow(model.RawSQL("SELECT COUNT(*) FROM log_entry")).
		Scan(&stat.Count); err != nil {
		return nil, err
	}

	if err := tx.GetEngine().
		QueryRow(model.RawSQL(
			"SELECT pg_total_relation_size('log_entry')")).
		Scan(&stat.EstimatedSize); err != nil {
		return nil, err
	}

	return &stat, nil
}
