// Copyright 2019 Orus.io. All rights reserved.

package storage

import (
	"testing"

	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
	"xbus.io/go-xbus/v3/api"

	"xbus.io/xbus/storage"
)

// PipelineStorageFactory is any factory of PipelineStorage
type PipelineStorageFactory func(bt testing.TB) (storage.PipelineStorage, func())

// PipelineStorageTests is a test suite for PipelineStorage implementations
func PipelineStorageTests(t *testing.T, factory PipelineStorageFactory) {
	t.Helper()

	t.Run("Save", func(t *testing.T) {
		ps, clean := factory(t)
		defer clean()

		id, err := ps.Save(storage.Pipeline{
			PipelineInfo: api.PipelineInfo{
				Name:    "test",
				Version: "0.1",
				Status:  api.PipelineInfo_DRAFT,
			},
			Graph: storage.Graph{
				Nodes: []storage.Node{{
					ID: "source",
				}},
			},
		})
		assert.Nil(t, err)

		p, ok := ps.Load(id)
		assert.True(t, ok)
		assert.Equal(t, "test", p.Name)
		assert.Equal(t, api.PipelineInfo_DRAFT, p.Status)
		assert.Equal(t, id, p.ID)

		assert.NoError(t, ps.SetStatus(id, api.PipelineInfo_ACTIVE))

		p, ok = ps.Load(id)
		assert.True(t, ok)
		assert.Equal(t, api.PipelineInfo_ACTIVE, p.Status)

		_, err = ps.Save(p)
		assert.Equal(t, storage.ErrInvalidPipelineStatus, err)

		p.ID = api.UUID{}
		_, err = ps.Save(p)
		assert.Equal(t, storage.ErrInvalidPipelineStatus, err)
	})

	t.Run("Update", func(t *testing.T) {
		ps, clean := factory(t)
		defer clean()

		p := storage.Pipeline{
			PipelineInfo: api.PipelineInfo{
				Name:    "test",
				Version: "0.1",
				Status:  api.PipelineInfo_DRAFT,
			},
			Graph: storage.Graph{
				Nodes: []storage.Node{
					{ID: "source"},
					{ID: "sink"},
				},
				Edges: []storage.Edge{
					storage.MustEdgeFromString("source->sink"),
				},
			},
		}

		id, err := ps.Save(p)
		assert.Nil(t, err)

		p, ok := ps.Load(id)
		require.True(t, ok)
		assert.Equal(t,
			[]storage.Edge{
				storage.MustEdgeFromString("source->sink"),
			},
			p.Graph.Edges)

		p.Graph.Edges = append(p.Graph.Edges, storage.MustEdgeFromString("source-> worker"))

		_, err = ps.Save(p)
		assert.Nil(t, err)

		p, ok = ps.Load(id)
		require.True(t, ok)
		assert.Equal(t,
			[]storage.Edge{
				storage.MustEdgeFromString("source->sink"),
				storage.MustEdgeFromString("source->worker"),
			},
			p.Graph.Edges)
	})
}
