// Copyright 2016 Orus.io. All rights reserved.

package api

const (
	// Msgbox is were envelopes are sent
	Msgbox = XbusAPIPrefix + "msgbox"

	// DefaultInput is the name of the default input
	DefaultInput = "default"
	// DefaultOutput is the name of the default output
	DefaultOutput = "default"
)

// EmitterMsgboxSubject returns the msgbox for an emitter
func EmitterMsgboxSubject(id UUID) string {
	return Msgbox + "." + id.String()
}

// ActorOutputSubject returns the subject for an actor output
func ActorOutputSubject(id UUID) string {
	return XbusAPIPrefix + "broker." + id.String() + ".output"
}

// ActorInputSubject returns the subject for an actor input
func ActorInputSubject(id UUID, output string) string {
	return Msgbox + "." + id.String() + ".input." + output
}

var (
	// EnvelopeMaxSize is an approximate maximum size of an envelope
	// A single Envelope should never exceed this size, or it could be
	// rejected by nats
	EnvelopeMaxSize = 750 * 1024 // 750K encrypted should be smaller than 1M
)

// NewEnvelopeAck creates a normal EnvelopeAck
func NewEnvelopeAck(id UUID, status EnvelopeAck_ReceptionStatus) *EnvelopeAck {
	return &EnvelopeAck{
		ID:     id,
		Status: status,
	}
}

// NewEnvelopeAckError prepare a EnvelopeAck from an error
func NewEnvelopeAckError(id UUID, err error) *EnvelopeAck {
	return &EnvelopeAck{
		ID:     id,
		Status: EnvelopeAck_ERROR,
		Reason: err.Error(),
	}
}

// ItemsAsChunks converts a slice of items into a slice of data chunks
func ItemsAsChunks(items []Item) [][]byte {
	var chunks = make([][]byte, len(items))

	for i, item := range items {
		chunks[i] = item
	}

	return chunks
}

// ChunksAsItems converts a slice of chunks into a slice of items
func ChunksAsItems(chunks [][]byte) []Item {
	var items = make([]Item, len(chunks))

	for i, chunk := range chunks {
		items[i] = chunk
	}

	return items
}

// EnvelopePositionStart returns the start position of an envelope
func EnvelopePositionStart(id UUID) EnvelopePosition {
	return EnvelopePosition{Start: true}
}

// Merge some envelope fragments into this envelope
func (envelope *Envelope) Merge(fragment ...*Envelope) error {
	for _, f := range fragment {
		envelope.Last = envelope.Last || f.Last
		// Merge EventIDs
		for _, evID := range f.EventIDs {
			var found bool
			for _, id := range envelope.EventIDs {
				if id == evID {
					found = true
					break
				}
			}
			if !found {
				envelope.EventIDs = append(envelope.EventIDs, evID)
			}
		}

		// Merge Events
		for _, ev := range f.Events {
			var found bool
			for iev, rev := range envelope.Events {
				if rev.ID == ev.ID {
					found = true
					rev.Items = append(rev.Items, ev.Items...)
					if ev.Checksum != 0 {
						rev.Checksum = ev.Checksum
					}
					if ev.Type != "" {
						rev.Type = ev.Type
					}
					if ev.ItemCount != 0 {
						rev.ItemCount = ev.ItemCount
					}
					envelope.Events[iev] = rev
					break
				}
			}
			if !found {
				envelope.Events = append(envelope.Events, ev)
			}
		}
	}
	return nil
}
